#include <medusa/Medusa.hpp>
#include <Eigen/SparseCore>
#include <Eigen/IterativeLinearSolvers>
#include <string>
#include <vector>
#include <algorithm>
#include <Eigen/SparseLU>


// *************************************************************************************
// Solution to Poisson equation in d-dimensional ball with Dirichlet boundary conditions.
// PROBLEM:
// \nabla ^2 u = - d \pi ^ 2  \prod _i \sin( \pi x_i) for i = 1, ..., d
//
// ANALYTIC SOLUTION:
// u(\bm x) = \prod _j \sin(\pi x_j)
//
// Author: Mitja Jančič, Institute Jozef Stefan
// Contact: Mitja.Jancic@ijs.si
//
// Date: 15.08.2019
// *************************************************************************************


using namespace mm;
using namespace std;

template <int dim, int r>
void analyse(const XML& conf, HDF& file) {

    typedef Vec<double, dim> vec; // dim-dimensional vector

    int n = conf.get<int>("case.n"); // discretization step
    int mon = conf.get<int>("num.monOrder"); // monomial order
    cout << "Working on mon order = " << mon << " and discretization n = " << n << endl;

    // Initialize timer
    Timer t;
    t.addCheckPoint("domain");
    prn("Getting domain.")

    // Domain definition and discretization
    const double origin = conf.get<double>("domain.origin"); // Ball origin
    const double radius = conf.get<double>("domain.radius"); // Ball radius
    double dx = 2 * radius / n;

    BallShape<vec> b1(origin, radius); // Define dim-dimensional ball
    DomainDiscretization<vec> domain = b1.discretizeBoundaryWithStep(dx); // Discretize boundary

    t.addCheckPoint("fill");
    prn("Adding nodes.")
    // Fill the interior using GeneralFill
    GeneralFill<vec> fill;
    const int fillSeed = conf.get<int>("num.fillSeed"); // Fill Seed
    fill.seed(fillSeed);
    auto fn = [=](const vec&) { return dx; }; // Lambda function in case you want to change density in domain
    fill(domain, fn);

    // Node indexes for interior, Dirichlet and Neumann Boundary condition
    // Dirichlet for x < 0.5 and Neumann for x >= 0.5
    Range<int> interior = domain.interior();
    Range<int> dir, neu;
    for (int i : domain.boundary()) {
        double x = domain.pos(i, 0);

        if (x < 0.5) {
            dir.push_back(i);
        } else {
            neu.push_back(i);
        }
    }

    Range<int> gh1(domain.size(), -1);
    // Ghost nodes to Neumann boundary nodes
    for (int i : neu) {
        gh1[i] = domain.addInternalNode(domain.pos(i) + dx * domain.normal(i), 2); // Choosen type 2
    }
    // Ghost nodes to Dirichlet boundary nodes
    for (int i : dir) {
        gh1[i] = domain.addInternalNode(domain.pos(i) + dx * domain.normal(i), 2); // Choosen type 2
    }

    int N = domain.size(); // Number of nodes in domain
    prn(N);

    // Support (stencil) nodes
    t.addCheckPoint("support");
    prn("Finding support.")

    int supportSize = conf.get<int>("num.support"); // Gets support size from xml
    if (supportSize <= 0) {
        // Change to default value if support in XML is negative
        auto allMon = Monomials<vec>(mon);
        supportSize = allMon.size() * 2;
        supportSize = max(supportSize, 2 * dim + 1);
    }
    auto allMonomials = Monomials<vec>(mon);
    prn(supportSize);

    // Find supports
    domain.findSupport(FindClosest(supportSize));

    t.addCheckPoint("shapes");
    prn("Calculating shapes.")
    // Basis functions
    Polyharmonic<double, r> ph;
    RBFFD<decltype(ph), vec, ScaleToFarthest> appr(ph, allMonomials);

    // Compute and store shapes (we only need laplacian and first derivatives)
    auto storage = domain.template computeShapes<sh::lap|sh::d1>(appr);

    // Declare global matrix and construct implicit operators
    t.addCheckPoint("matrix");
    prn("Filling global matrix.")
    Eigen::SparseMatrix<double, Eigen::RowMajor> M(N, N);
    Eigen::VectorXd rhs(N);
    rhs.setZero(); // Set RHS to zero

    auto op = storage.implicitOperators(M, rhs);
    M.reserve(storage.supportSizes());

    // Lambda function returning sin product
    auto getSinProduct = [](const auto& pos, int k = -1) {
        int d = pos.size(); // Space dimensionality -> not needed, could also simply use dim from template

        double sinProduct = 1.0;
        for (int j = 0; j < d; j++) {
            auto x = pos[j];

            if (k == j) {
                sinProduct *= std::cos(PI * x);
            } else {
                sinProduct *= std::sin(PI * x);
            }
        }

        return sinProduct;
    };


    // Solve Poisson's equation for interior
    for (int i : interior) {
        double sinProduct = getSinProduct(domain.pos(i));

        // Equation
        op.lap(i) = -dim * PI * PI * sinProduct;
    }

    // Enforce Dirichlet boundary condition
    for (int i : dir) {
        double sinProduct = getSinProduct(domain.pos(i));

        op.value(i) = sinProduct;

        // Equation holds also on the boundary, write it in the row corresponding to the ghost node.
        op.lap(i, gh1[i]) = -dim * PI * PI * sinProduct;
    }

    // Enforce Neumann boundary condition
    for (int i : neu) {
        double boundaryVal = 0.0;

        //Iterate over all terms
        for (int j = 0; j < dim; j++){
            // Compute term
            double sinProduct = getSinProduct(domain.pos(i), j);

            boundaryVal += PI * sinProduct * domain.normal(i)(j);
        }

        op.neumann(i, domain.normal(i)) = boundaryVal;  // Neumann

        //Equation holds also on the boundary, write it in the row corresponding to the ghost node.
        double sinProduct = getSinProduct(domain.pos(i));
        op.lap(i, gh1[i]) = -dim * PI * PI * sinProduct;
    }

    // Solve system
    t.addCheckPoint("compute");
    prn("Computing.")
    Eigen::BiCGSTAB<decltype(M), Eigen::IncompleteLUT<double>> solver;

    // Preconditioner settings
    if (conf.get<int>("solver.usePrecon") > 0) {
        double fillFactor = conf.get<double>("solver.fill");
        double droptol = conf.get<double>("solver.droptol");
        int maxiter = conf.get<int>("solver.maxiter");
        double globtol = conf.get<double>("solver.globtol");

        solver.preconditioner().setFillfactor(fillFactor);
        solver.preconditioner().setDroptol(droptol);
        solver.setMaxIterations(maxiter);
        solver.setTolerance(globtol);
    }

    solver.compute(M);
    t.addCheckPoint("solve");
    prn("Solving.")
    ScalarFieldd u = solver.solve(rhs);

    prn(solver.iterations()) // Print solver iterations
    prn(solver.error()) // print solver error

    // Write solution to file
    t.addCheckPoint("postprocess");
    prn("Saving to file.")

    file.setGroupName(format("Discretization_dim%d_r%d_n%.4d_mon%d_sup%d",
                             dim, r, n, mon, supportSize)); // creates new group
    file.atomic().writeDouble2DArray("pos", domain.positions()); // positions
    file.atomic().writeDoubleArray("sol", u); // solution
    file.atomic().writeIntArray("type", domain.types()); // node type
    file.atomic().writeSparseMatrix("mat", M); // sparse matrix M
    file.atomic().writeIntAttribute("solverIterations", solver.iterations()); // solver iterations
    file.atomic().writeDoubleAttribute("solverError", solver.error()); // solver error

    // End Timer
    t.addCheckPoint("end");
    file.atomic().writeTimer("timer", t);
}

int main(int argc, char *argv[]) {

    // Initialize timer
    Timer t;
    t.addCheckPoint("begin");

    // Read XML
    XML conf(argv[1]);
    int dim = conf.get<int>("domain.dim");

    //h5 file to store the solution inside
    string fileName = conf.get<string>("out.fileName");
    cout << "Creating output file ..." << endl;
    string output_file = format("Ball_dim%d_%s.h5", dim, fileName);
    HDF file(output_file, HDF::DESTROY);

    // Write xml to h5
    file.writeXML("conf", conf);
    file.close();

    // Compute
    if (dim == 1) {
        analyse<1, 3>(conf, file);
    } else if (dim == 2) {
        analyse<2, 3>(conf, file);
    } else if (dim == 3) {
        analyse<3, 3>(conf, file);
    } else  {
        std::cout << "No definition for dimensions greater than 3." << std::endl;
    }

    // End
    t.addCheckPoint("end");
    prn(t.duration("begin", "end"));

    std::cout << "Calculations saved to " << output_file << std::endl;
    return 0;
}
